from typing import Dict

from .exceptions import *
from .lifecycle import *


class ProvideTarget:
    # This is a class to store the dependencies definitions within a module.
    def __init__(self, callable: callable, provides: str, requires: [str]):
        self.callable = callable
        self.provides = provides
        self.requires = requires


class InvokeTarget:
    # This is a class to store the invoke definitions within a module.
    def __init__(self, callable: callable, requires: [str]):
        self.callable = callable
        self.requires = requires


class Provider:
    # This is a class that is used to build advanced dependencies definitions.
    def __init__(self, provider: object, name: str | None = None, group: bool = False):
        self.provider = provider
        self.name = name
        if name is None:
            self.name = provider
        self.group = group


class Option:
    # This is the base class for the functional options pattern.
    def apply(self, mod: 'Module'):
        pass


class Module(Option):
    # This is a class that is used to store all the dependencies and invoke targets,
    # and also holds the lifecycle object and the built dependencies.
    def __init__(self, *args):
        self.lifecycle = Lifecycle()
        self._provides: Dict[str, ProvideTarget] = {}
        self._invokes: [InvokeTarget] = []

        for option in args:
            if not isinstance(option, Option):
                raise PyDITypeError(
                    "App constructor arguments must be of type Option")
            option.apply(self)

    def apply(self, mod: 'Module'):
        mod.add_provides(self._provides)
        mod.add_invokes(self._invokes)

    def register_container(self, container: Dict[str, ProvideTarget]) -> None:
        self.container = container

    def set_provides(self, provides: Dict[str, ProvideTarget]) -> None:
        self._provides = provides

    def get_provides(self) -> Dict[str, ProvideTarget]:
        return self._provides

    def build_dependencies(self, target: ProvideTarget | InvokeTarget) -> []:
        inject = []
        for require in target.requires:
            if require == Lifecycle:
                inject.append(self.lifecycle)
                continue

            key = require
            if isinstance(require, Provider):
                key = require.name

            if key not in self._provides:
                raise MissingDependencyError("Cannot find dependency", key)

            if key in self.container:
                if isinstance(require, Provider) and not issubclass(self.container[key].__class__, require.provider):
                    raise DependencyTypeError("Dependency is of wrong type", key,
                                              self.container[key].__class__, "but should be of type", require.provider)
                inject.append(self.container[key])
            else:
                provider = self.build_provider(self._provides[key])
                if isinstance(require, Provider):
                    if not require.group and not issubclass(provider.__class__, require.provider):
                        raise DependencyTypeError("Dependency is of wrong type", key,
                                                  provider.__class__, "but should be of type", require.provider)
                    if require.group and not issubclass(provider[0].__class__, require.provider):
                        raise DependencyTypeError("Dependency is of wrong type", key,
                                                  provider[0].__class__, "but should be of type", require.group)
                self.container[key] = provider
                inject.append(provider)

        return inject

    def build_provider(self, target: ProvideTarget | list[ProvideTarget]) -> object | list[object]:
        if isinstance(target, ProvideTarget):
            inject = self.build_dependencies(target)
            try:
                return target.callable(*inject)
            except TypeError as e:
                if "missing" in str(e):
                    raise MissingHintError(
                        str(e), "verify if all hints are set")
                raise e

        elements = []
        for element in target:
            if not isinstance(element, ProvideTarget):
                raise PyDITypeError(
                    "Provide target list element must be of type ProvideTarget")
            inject = self.build_dependencies(element)
            try:
                elements.append(element.callable(*inject))
            except TypeError as e:
                if "missing" in str(e):
                    raise MissingHintError(
                        str(e), "verify if all hints are set")
                raise e
        return elements

    def add_provides(self, targets: dict):
        if not isinstance(targets, dict):
            raise PyDITypeError("Provide targets must be a dict")
        for key, value in targets.items():
            if not isinstance(value, ProvideTarget) and not isinstance(value, list):
                raise PyDITypeError(
                    "Provide target must be of type ProvideTarget or list of ProvideTarget")
            if isinstance(value, list):
                for target in value:
                    if not isinstance(target, ProvideTarget):
                        raise PyDITypeError(
                            "Provide target list element must be of type ProvideTarget")

        self._provides = self._provides | targets

    def add_invokes(self, targets: [InvokeTarget]):
        for target in targets:
            if not isinstance(target, InvokeTarget):
                raise PyDITypeError(
                    "Invoke target must be of type InvokeTarget")
        self._invokes.extend(targets)



import unittest

from src.dinjections import *


class TestClass1:
    pass


class TestClass2:
    pass


class TestClass3:
    pass

    def run(self):
        pass


def new_test_class_3(p: Provider(TestClass1, "t1"), p2: Provider(TestClass1, "t2"), r: TestClass2) -> TestClass3:
    return TestClass3()


def register_hooks(l: Lifecycle, t: TestClass3):
    l.append_hook(Hook(
        on_start=lambda: {
            t.run()
        },
    ))


def register_hooks_no_hint(l: Lifecycle, t):
    l.append_hook(Hook(
        on_start=lambda: {
            t.run()
        },
    ))


class TestApp(unittest.TestCase):

    def test_app(self):
        try:
            app = App(
                Provide(
                    # test named provider
                    Provider(TestClass1, "t1"),
                    Provider(TestClass1, "t2"),
                    TestClass2,
                    new_test_class_3,
                ),
                Invoke(
                    register_hooks,
                ))
            app.run()
        except Exception as e:
            self.fail("Exception: " + str(e))
        self.assertTrue(True)

    def test_missing_dependency(self):
        exc = None
        try:
            app = App(
                Provide(),
                Invoke(
                    register_hooks,
                ))

            app.run()
        except Exception as e:
            exc = e

        self.assertIsInstance(exc, MissingDependencyError)

    def test_dependency_type_error(self):
        exc = None
        try:
            provides = Provide(
                Provider(TestClass1, "t1"),
                Provider(TestClass2, "t2"),
                TestClass2,
                new_test_class_3,
            )
            app = App(
                provides,
                Invoke(
                    register_hooks,
                ))

            app.run()
        except Exception as e:
            exc = e

        self.assertIsInstance(exc, DependencyTypeError)

    def test_missing_hint_error(self):
        def new_test_class_3_no_hint(p: Provider(TestClass1, "t1"), p2: Provider(TestClass1, "t2"), r: TestClass2):
            return TestClass3()

        exc = None
        try:
            provides = Provide(
                Provider(TestClass1, "t1"),
                Provider(TestClass1, "t2"),
                TestClass2,
                new_test_class_3_no_hint,
            )
            app = App(
                provides,
                Invoke(
                    register_hooks,
                ))

            app.run()
        except Exception as e:
            exc = e

        self.assertIsInstance(exc, MissingHintError)

    def test_missing_hint_error_input_invoke(self):
        exc = None
        try:
            provides = Provide(
                Provider(TestClass1, "t1"),
                Provider(TestClass1, "t2"),
                TestClass2,
                new_test_class_3,
            )
            app = App(
                provides,
                Invoke(
                    register_hooks_no_hint,
                ))

            app.run()
        except Exception as e:
            exc = e

        self.assertIsInstance(exc, MissingHintError)

    def test_missing_hint_error_input_provide(self):
        def new_test_class_3_no_hint_input(p: Provider(TestClass1, "t1"), p2: Provider(TestClass1, "t2"), r) -> TestClass3:
            return TestClass3()

        exc = None
        try:
            provides = Provide(
                Provider(TestClass1, "t1"),
                Provider(TestClass1, "t2"),
                TestClass2,
                new_test_class_3_no_hint_input,
            )
            app = App(
                provides,
                Invoke(
                    register_hooks,
                ))

            app.run()
        except Exception as e:
            exc = e

        self.assertIsInstance(exc, MissingHintError)

    def test_group(self):
        def new_test_class_3_group(p: Provider(TestClass1, group=True), r: TestClass2) -> TestClass3:
            return TestClass3()

        try:
            app = App(
                Provide(
                    Provider(TestClass1, group=True),
                    Provider(TestClass1, group=True),
                    TestClass2,
                    new_test_class_3_group,
                ),
                Invoke(
                    register_hooks,
                ))
            app.run()
        except Exception as e:
            self.fail("Exception: " + str(e))
        self.assertTrue(True)

    def test_modularity(self):
        m1 = Module(Provide(
            Provider(TestClass1, name="t1"),
            Provider(TestClass1, name="t2"),
        ))
        m2 = Module(Provide(
            TestClass2,
            new_test_class_3,
        ),
            Invoke(
                register_hooks,
        ))
        try:
            app = App(Module(m1, m2))
            app.run()
        except Exception as e:
            self.fail("Exception: " + str(e))
        self.assertTrue(True)

    def test_callable_with_provider(self):
        def new_test_class_3_named(p: TestClass1) -> Provider(TestClass3, name="t3"):
            return TestClass3()
        def register_hooks_named(l: Lifecycle, t: Provider(TestClass3, name="t3")):
            l.append_hook(Hook(
                on_start=lambda: {
                    t.run()
                },
            ))
        try:
            app = App(Module(Provide(
                TestClass1,
                new_test_class_3_named,
            ),
                Invoke(
                register_hooks_named,
            )))
            app.run()
        except Exception as e:
            self.fail("Exception: " + str(e))
        self.assertTrue(True)

    def test_cyclic_dependency(self):
        def new_test_class_3_cyclic(p: TestClass3) -> TestClass3:
            return TestClass3()

        exc = None
        try:
            app = App(
                Provide(new_test_class_3_cyclic),
                Invoke(
                    register_hooks,
                ))

            app.run()
        except Exception as e:
            exc = e

        self.assertIsInstance(exc, RecursionError)

    def test_inherited(self):
        class TestClass4(TestClass3):
            pass

        def register_hooks(l: Lifecycle, t: TestClass3):
            l.append_hook(Hook(
                on_start=lambda: {
                    t.run()
                },
            ))

        try:
            app = App(
                Provide(
                    Provider(TestClass4, name=TestClass3),
                ),
                Invoke(
                    register_hooks,
                ))

            app.run()
        except Exception as e:
            self.fail("Exception: " + str(e))
        self.assertTrue(True)


if __name__ == '__main__':
    unittest.main()
